import scala.io.Source
import scala.io
import java.io.File
import scala.util.parsing.combinator._

sealed class AppliedPiTokenParsers extends JavaTokenParsers with PackratParsers {
  override protected val whiteSpace = """(\s|%.*)+""".r
  def ignore (s:String) : Unit = { () }
  def constantToken (t : String) : PackratParser[Unit] =
    literal(t) ^^ ignore
  val integer : PackratParser[Int] = 
    wholeNumber ^^ (s => s.toInt)
  val identifier : PackratParser[String] = 
    regex("""[a-zA-Z][a-zA-Z0-9]*""".r)
}

 
sealed class TermParsers extends AppliedPiTokenParsers {
  val expr : PackratParser[Term] =
    orexpr~("::"~>expr) ^^ {case h~TermList(x) => new TermList(Append(h,x))} | "[]" ^^^ (TermList(EmptyList)) | orexpr
 
  val orexpr : PackratParser[Term] = 
    orexpr~("\\/"~>andexpr) ^^ {case x~y => new Or(x,y)} | andexpr
  
  val andexpr : PackratParser[Term] =
    andexpr~("/\\"~>supexpr) ^^ {case x~y => new And(x,y)} | supexpr

  val supexpr : PackratParser[Term] =
    supexpr~(">"~>eqexpr) ^^ {case x~y => new Sup(x,y)} | eqexpr

  val eqexpr : PackratParser[Term] =
    eqexpr~("="~>termexpr) ^^ {case x~y => new Eq(x,y)} | termexpr

  val integerTerm : PackratParser[Term] =
    integer ^^ (s => new Integer(s))
  val variable : PackratParser[Term] =
    identifier ^^ (s => new Var(s))
  val count : PackratParser[Term] =
    "count"~"("~expr~")" ^^ {case "count"~"("~t~")" => new Count(t)}
  val not : PackratParser[Term] =
    "not"~("("~>expr<~")") ^^ {case "not"~t => new Not(t)}
  val pi1 : PackratParser[Term] =
    "pi1"~("("~>expr<~")") ^^ {case "pi1"~t => new Pi1(t)}
  val pi2 : PackratParser[Term] =
    "pi2"~("("~>expr<~")") ^^ {case "pi2"~t => new Pi2(t)}
  val pk : PackratParser[Term] =
    "pk"~("("~>expr<~")") ^^ {case "pk"~t => new Pk(t)}
  val sk : PackratParser[Term] =
    "sk"~("("~>expr<~")") ^^ {case "sk"~t => new Sk(t)}
  val pair : PackratParser[Term] =
    "pair"~("("~>expr<~",")~(expr<~")") ^^ {case "pair"~left~right => new Pair(left,right)}
  val dec : PackratParser[Term] =
    "dec"~("("~>expr<~",")~(expr<~")") ^^ {case "dec"~left~right => new Dec(left,right)}
  val enc : PackratParser[Term] =
    "enc"~("("~>expr<~",")~(expr<~")") ^^ {case "enc"~left~right => new Enc(left,right)}

  val termexpr : PackratParser[Term] =
    integerTerm |  count | not | pi1 | pi2 | enc | dec | pk | sk | pair | variable | "("~>expr<~")"

  val term : PackratParser[Term] =
    expr

}


sealed class BasicProcessParsers extends TermParsers {
  def nullProcess : PackratParser[BasicProcess] =
    "0" ^^^ (NullProcess)
  def outProcess : PackratParser[BasicProcess] =
    "out"~"("~identifier~","~term~")"~"."~basicProcess ^^ {case "out"~"("~id~","~t~")"~"."~basic => new Out(ChannelManager.getChannel(id), t, basic)}
  def inProcess : PackratParser[BasicProcess] =
    "in"~"("~identifier~","~identifier~")"~"."~basicProcess ^^ {case "in"~"("~id~","~v~")"~"."~basic => new In(ChannelManager.getChannel(id), new Var(v), basic)}
  def multiIn : PackratParser[BasicProcess] =
    "in"~"^"~integer~"("~identifier~","~identifier~"->"~term~"as"~identifier~")"~"."~basicProcess ^^ {
      case "in"~"^"~k~"("~chan~","~x~"->"~u~"as"~y~")"~"."~p => new MultiIn(ChannelManager.getChannel(chan), k, new Var(x), u, new Var(y), p)
    }
  def ifThenElse : Parser[BasicProcess] =
    "if"~term~"then"~basicProcess~"else"~basicProcess ^^ {case "if"~v~"then"~p1~"else"~p2 => new IfThenElse(v, p1, p2)}
  def newConstant : PackratParser[BasicProcess] =
    ("new"~identifier~"."~basicProcess ^^ {
	case "new"~s~"."~p => new NewConstant(new Var(s), p)
      }) | ("("~"new"~identifier~")"~"."~basicProcess ^^ {
	case "("~"new"~s~")"~"."~p => new NewConstant(new Var(s), p)
      })

  def parenthesisedBasicProcess : PackratParser[BasicProcess] =
    "("~basicProcess~")" ^^ {case "("~p~")" => p}

  def basicProcess : PackratParser[BasicProcess] =
    nullProcess | outProcess | inProcess | multiIn | ifThenElse | newConstant | parenthesisedBasicProcess
}


object ProcessParsers extends BasicProcessParsers {
  def replication : PackratParser[Process] =
    (basicProcess~"^"~integer ^^ {case p~"^"~n => new Replication(p,n)}) | basicProcess ^^ (p => new Replication(p,1))

  def process : PackratParser[Process] =
    (replication~"||"~process ^^ {case p1~"||"~p2 => new Parallel(p1,p2)}) | replication

  def getProcess(r : java.io.FileReader) : Process = {
    parseAll(process,r) match {
      case Success(result,_) => result
      case failure : NoSuccess =>  scala.sys.error(failure.msg)
    }
  }
}

object InterpretedTermParsers extends AppliedPiTokenParsers {
  val expr : PackratParser[Term] =
    termexpr~("::"~>expr) ^^ {case h~TermList(x) => new TermList(Append(h,x))} | "[]" ^^^ (TermList(EmptyList)) | termexpr
 
  val integerTerm : PackratParser[Term] =
    integer ^^ (s => new Integer(s))
  val string : PackratParser[Term] =
    identifier ^^ (s => new TString(s))
  val count : PackratParser[Term] =
    "count"~"("~expr~")" ^^ {case "count"~"("~t~")" => new Count(t)}
  val not : PackratParser[Term] =
    "not"~("("~>expr<~")") ^^ {case "not"~t => new Not(t)}
  val pi1 : PackratParser[Term] =
    "pi1"~("("~>expr<~")") ^^ {case "pi1"~t => new Pi1(t)}
  val pi2 : PackratParser[Term] =
    "pi2"~("("~>expr<~")") ^^ {case "pi2"~t => new Pi2(t)}
  val pk : PackratParser[Term] =
    "pk"~("("~>expr<~")") ^^ {case "pk"~t => new Pk(t)}
  val sk : PackratParser[Term] =
    "sk"~("("~>expr<~")") ^^ {case "sk"~t => new Sk(t)}
  val pair : PackratParser[Term] =
    "pair"~("("~>expr<~",")~(expr<~")") ^^ {case "pair"~left~right => new Pair(left,right)}
  val dec : PackratParser[Term] =
    "dec"~("("~>expr<~",")~(expr<~")") ^^ {case "dec"~left~right => new Dec(left,right)}
  val enc : PackratParser[Term] =
    "enc"~("("~>expr<~",")~(expr<~")") ^^ {case "enc"~left~right => new Enc(left,right)}

  val termexpr : PackratParser[Term] =
    integerTerm |  count | not | pi1 | pi2 | enc | dec | pk | sk | pair | string | "("~>expr<~")"

  val term : PackratParser[Term] =
    expr

  def getTerm(s : String) : Term =
    parseAll(term,s) match {
      case Success(result,_) => result
      case failure : NoSuccess =>  scala.sys.error(failure.msg)
    }
}
