
object rnd extends scala.util.Random

/******************************************************************************/
/*                    BASIC PROCESSES DEFINITIONS                             */
/******************************************************************************/


sealed trait BasicProcess extends Thread {
  def evaluate(env : PartialFunction[String,Term]) : Unit
  override def run() {
    val nada : PartialFunction[String,Term] = Map()
    this.evaluate(nada)
  }
  def copy() : BasicProcess
}

case object NullProcess extends BasicProcess {
  def evaluate(env : PartialFunction[String,Term]){
  }
  def execute(){
  }
  override def toString() : String = {
    "0"
  }
  def copy() : BasicProcess = {
    NullProcess
  }
}
case class NewConstant(s : Var, p : BasicProcess) extends BasicProcess {
  def evaluate(env : PartialFunction[String,Term]){
    val b = rnd.nextBoolean()
    val nextMatching : PartialFunction[String,Term] = {
      case x if (x == s) => Integer(if (b) 1 else 0)
    }
    p.evaluate(nextMatching orElse env)
  }
  override def toString() : String = {
    "(new "+s.toString()+")."+p.toString()
  }
  def copy() : BasicProcess = {
    new NewConstant(s,p.copy())
  }
}

case class Out(c : Channel, t : Term, p : BasicProcess) extends BasicProcess {
  def evaluate(env : PartialFunction[String,Term]){
    val toSend = t.evaluate(env).interpret()
    c.send(toSend);
    p.evaluate(env)
  }
  override def toString() : String = {
    "out("+c.toString()+","+t.toString()+")."+p.toString()
  }
  def copy() : BasicProcess = {
    new Out(c,t,p.copy())
  }
}

case class In(c : Channel, v : Var, p : BasicProcess) extends BasicProcess {
  def evaluate(env : PartialFunction[String,Term]){
    val received = c.receive()
    val t = InterpretedTermParsers.getTerm(received)
    val nextBinding : PartialFunction[String,Term] = {
      case v.id => t
    }
    p.evaluate(nextBinding orElse env)
  }
  override def toString() : String = {
    "in("+c.toString()+","+v.toString()+")."+p.toString()
  }
  def copy() : BasicProcess = {
    new In(c,v,p.copy())
  }
}

case class MultiIn(c : Channel, k : Int, x : Var, u : Term, y : Var, p : BasicProcess) extends BasicProcess {
  def evaluate(env : PartialFunction[String,Term]){
    var l : TList = EmptyList
    for (i <- 1 to k){
      val received = c.receive()
      val t = InterpretedTermParsers.getTerm(received)
      l = Append(u.evaluate({case x.id => t}),l)
    }
    val nextBinding : PartialFunction[String,Term] = {
      case y.id => TermList(l)
    }
    p.evaluate(nextBinding orElse env)
  }
  override def toString() : String = {
    "in^"+k.toString()+"("+c.toString()+","+x.toString()+"->"+u.toString()+" as "+y.toString()+")."+p.toString()
  }
  def copy() : BasicProcess = {
    new MultiIn(c,k,x,u,y,p.copy())
  }
}

case class IfThenElse(c : Term, pThen : BasicProcess, pElse : BasicProcess) extends BasicProcess {
  def evaluate(env : PartialFunction[String,Term]){
    val condition = c.evaluate(env)
    condition match{
      case Integer(i) if (i != 0) => pThen.evaluate(env)
      case Integer(i) if (i == 0) => pElse.evaluate(env)
      case _ => throw new EvaluationError("condition did not evaluate to an integer")
    }
  }
  override def toString() : String = {
    "if ("+c.toString()+") then ("+pThen.toString()+") else ("+pElse.toString()+")"
  }
  def copy() : BasicProcess = {
    new IfThenElse(c,pThen.copy(),pElse.copy())
  }
}

/******************************************************************************/
/*                    GENERAL PROCESSES DEFINITIONS                           */
/******************************************************************************/


sealed abstract class Process {
  def execute() : Unit
}

case object Null extends Process {
  def execute(){
  }
  override def toString() : String = {
    "0"
  }
}
case class Parallel(p1 : Process , p2 : Process) extends Process {
  def execute() {
    p1.execute();
    p2.execute()
  }
  override def toString() : String = {
    p1.toString()+" || "+p2.toString()
  }
}
case class Replication(p : BasicProcess, k : Int) extends Process {
  def execute() {
    for (i <- 1 to k){
      val toStart = p.copy()
      toStart.start()
    }
  }
  override def toString() : String = {
    "("+p.toString()+")"+"^"+k.toString()
  }
}
