/***********************************/
/*DEFINING TERMS AND LISTS OF TERMS*/
/***********************************/

case class InterpretationError(msg : String) extends Exception
case class EvaluationError(msg : String) extends Exception

sealed abstract class Term {
  def evaluate(env : PartialFunction[String,Term]) : Term
  def interpret() : String //A term should always be evaluated before being interpreted
}

sealed abstract class TList {
  def evaluate (env : PartialFunction[String,Term]) : TList
  def interpret () : String
  def count () : Int
}

case class Integer(v : Int) extends Term {
  def evaluate (env : PartialFunction[String,Term]) : Term = {
    this
  }
  def interpret () : String = {
    v.toString
  }
  override def toString () : String = {
    v.toString
  }
}

case class Var(id : String) extends Term {
  def evaluate (env : PartialFunction[String,Term]) : Term = {
    env.applyOrElse (this.id,(x:String) => TString(x))
  }
  def interpret () : String = {
    id
  }
  override def toString () : String = {
    id
  }
}

case class TermList(var l : TList) extends Term {
  def evaluate (env : PartialFunction[String,Term]) : Term = {
     TermList(l.evaluate(env))
  }
  def interpret () : String = {
    "("+(l.interpret())+")"
  }
  override def toString = interpret
}

case class Count(l : Term) extends Term {
  def evaluate (env : PartialFunction[String,Term]) : Term = {
    val countable = l.evaluate(env)
    countable match {
      case TermList(list) =>  Integer(list.count())
      case _ => throw EvaluationError("You are trying to count a non list")
    }
  }
  def interpret () : String = {
    throw  InterpretationError("You are trying to interpret a non evaluated count")
  }
  override def toString () : String = {
    "count("+l.toString()+")"
  }
}

case class Sup(lhs : Term, rhs : Term) extends Term {
  def evaluate(env : PartialFunction[String,Term]) : Term = {
    val big = lhs.evaluate(env)
    val small = rhs.evaluate(env)
    (big,small) match {
      case (Integer(b),Integer(s)) =>  Integer(if (b > s) 1 else 0)
      case _ => throw  EvaluationError("You are trying to compare non integer terms")
    }
  }
  def interpret() : String = {
    throw new InterpretationError("You are trying to interpret a non evaluated comparison")
  }
  override def toString() : String = {
    "("+lhs.toString() + ">" + rhs.toString()+")"
  }
}

case class Eq(lhs : Term, rhs : Term) extends Term {
  def evaluate(env : PartialFunction[String,Term]) : Term = {
    val left = lhs.evaluate(env)
    val right = rhs.evaluate(env)
    if (left == right) Integer(1) else Integer(0)
  }
  def interpret() : String = {
    throw new InterpretationError("You are trying to interpret a non evaluated equality")
  }
  override def toString() : String = {
    "("+lhs.toString() + "=" + rhs.toString()+")"
  }
}

case class And(lhs : Term, rhs : Term) extends Term {
  def evaluate(env : PartialFunction[String,Term]) : Term = {
    val left = lhs.evaluate(env)
    val right = rhs.evaluate(env)
    (left,right) match {
      case (Integer(l),Integer(r)) =>  Integer(if ((l != 0)&&(r != 0)) 1 else 0)
      case _ => throw new EvaluationError("You are trying to compare non boolean terms")
    }
  }
  def interpret() : String = {
    throw new InterpretationError("You are trying to interpret a non evaluated and")
  }
  override def toString() : String = {
    "("+lhs.toString() + "/\\" + rhs.toString()+")"
  }
}

case class Or(lhs : Term, rhs : Term) extends Term {
  def evaluate(env : PartialFunction[String,Term]) : Term = {
    val left = lhs.evaluate(env)
    val right = rhs.evaluate(env)
    (left,right) match {
      case (Integer(l),Integer(r)) =>  Integer(if ((l != 0)||(r != 0)) 1 else 0)
      case _ => throw new EvaluationError("You are trying to compare non boolean terms")
    }
  }
  def interpret() : String = {
    throw new InterpretationError("You are trying to interpret a non evaluated or")
  }
  override def toString() : String = {
    "("+lhs.toString() + "\\/" + rhs.toString()+")"
  }
}

case class Not(v : Term) extends Term {
  def evaluate(env : PartialFunction[String,Term]) : Term = {
    val value = v.evaluate(env)
    value match {
      case Integer(l) =>  Integer(if (l == 0) 1 else 0)
      case _ => throw new EvaluationError("You are trying to negate a non boolean term")
    }
  }
  def interpret() : String = {
    throw new InterpretationError("You are trying to interpret a non evaluated not")
  }
  override def toString() : String = {
    "not("+v.toString()+")"
  }
}

case class Pair(t1 : Term, t2 : Term) extends Term {
  def evaluate(env : PartialFunction[String,Term]) : Term = {
     Pair(t1.evaluate(env), t2.evaluate(env))
  }
  def interpret() : String = {
    "pair("+t1.interpret()+","+t2.interpret()+")"
  }
  override def toString() : String = {
    "pair("+t1.toString()+","+t2.toString()+")"
  }
}

case class Pi1(t : Term) extends Term {
  def evaluate(env : PartialFunction[String,Term]) : Term = {
    val p = t.evaluate(env)
    p match {
      case Pair(u,v) => u
      case _ => throw new EvaluationError("You are trying to project a non pair")
    }
  }
  def interpret() : String = {
    throw new InterpretationError("You are trying to interpret a non evaluated pi1")
  }
  override def toString() : String = {
    "pi1("+t.toString()+")"
  }
}

case class Pi2(t : Term) extends Term {
  def evaluate(env : PartialFunction[String,Term]) : Term = {
    val p = t.evaluate(env)
    p match {
      case Pair(u,v) => v
      case _ => throw new EvaluationError("You are trying to project a non pair")
    }
  }
  def interpret() : String = {
    throw new InterpretationError("You are trying to interpret a non evaluated pi2")
  }
  override def toString() : String = {
    "pi2("+t.toString()+")"
  }
}

case class Enc(m : Term, k : Term) extends Term {
  def evaluate(env : PartialFunction[String,Term]) : Term = {
     Enc(m.evaluate(env), k.evaluate(env))
  }
  def interpret() : String = {
    "enc("+m.interpret()+","+k.interpret()+")"
  }
  override def toString() : String = {
    "enc("+m.toString()+","+k.toString()+")"
  }
}

case class Dec(c : Term, k : Term) extends Term {
  def evaluate(env : PartialFunction[String,Term]) : Term = {
    (c.evaluate(env),k.evaluate(env)) match {
      case (Enc(u,Pk(i)),Sk(j)) => 
	if (i==j){
	  u
	}
	else {
	  throw new EvaluationError("You are trying to decrypt with a wrong key")
	}
      case _ => throw new EvaluationError("You are trying to decrypt an invalid encryption")
    }
  }
  def interpret() : String = {
    throw new InterpretationError("You are trying to interpret a non evaluated dec")
  }
  override def toString() : String = {
    "dec("+c.toString()+","+k.toString()+")"
  }
}

case class Pk(v : Term) extends Term {
  def evaluate(env : PartialFunction[String,Term]) : Term = {
    v.evaluate(env) match {
      case Integer(i) => Pk(Integer(i))
      case _ => throw new EvaluationError("You are trying to build a Pk from a non integer term")
    }
  }
  def interpret() : String = {
    "pk("+v.interpret()+")"
  }
  override def toString() : String = {
    "pk("+v.toString()+")"
  }
}

case class Sk(v : Term) extends Term {
  def evaluate(env : PartialFunction[String,Term]) : Term = {
    v.evaluate(env) match {
      case Integer(i) => Sk(Integer(i))
      case _ => throw new EvaluationError("You are trying to build a Sk from a non integer term")
    }
  }
  def interpret() : String = {
    "sk("+v.interpret()+")"
  }
  override def toString() : String = {
    "sk("+v.toString()+")"
  }
}

case class TString(s : String) extends Term {
  def evaluate(env : PartialFunction[String,Term]) : Term = {
    this
  }
  def interpret() : String = {
    s
  }
  override def toString() : String = {
    s
  }
}


case object EmptyList extends TList{
  def evaluate (env : PartialFunction[String,Term]) : TList = {
    this
  }
  def interpret () : String = {
    "[]"
  }
  def count() : Int = {
    0
  }
  override def toString () : String = {
    "[]"
  }
}

case class Append(t : Term, l : TList) extends TList {
  def evaluate (env : PartialFunction[String,Term]) : TList = {
     Append(t.evaluate(env), l.evaluate(env))
  }
  def interpret() : String = {
    t.interpret()+"::"+l.interpret()
  }
  def count() : Int = {
    val k = l.count()
    t match {
      case Integer(0) => k
      case Integer(_) => k+1
      case _ => k
    }
  }
  override def toString() : String = {
    t.toString()+"::"+l.toString()
  }
}
